@@ -1,9 +1,10 @@
 /* SASL server API implementation
  * Rob Siemborski
  * Tim Martin
+ * $Id: server.c,v 1.176 2011/09/01 16:33:10 mel Exp $
  */
 /* 
- * Copyright (c) 1998-2016 Carnegie Mellon University.  All rights reserved.
+ * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -21,13 +22,12 @@
  *    endorse or promote products derived from this software without
  *    prior written permission. For permission or any other legal
  *    details, please contact  
+ *      Office of Technology Transfer
  *      Carnegie Mellon University
- *      Center for Technology Transfer and Enterprise Creation
- *      4615 Forbes Avenue
- *      Suite 302
- *      Pittsburgh, PA  15213
- *      (412) 268-7393, fax: (412) 268-7395
- *      innovation@andrew.cmu.edu
+ *      5000 Forbes Avenue
+ *      Pittsburgh, PA  15213-3890
+ *      (412) 268-4387, fax: (412) 268-7395
+ *      tech-transfer@andrew.cmu.edu
  *
  * 4. Redistributions of any form whatsoever must retain the following
  *    acknowledgment:
@@ -90,7 +90,7 @@ static int _sasl_checkpass(sasl_conn_t *conn,
 
 static mech_list_t *mechlist = NULL; /* global var which holds the list */
 
-static sasl_global_callbacks_t global_callbacks;
+sasl_global_callbacks_t global_callbacks;
 
 /* set the password for a user
  *  conn        -- SASL connection
@@ -411,9 +411,6 @@ static int mech_compare(const sasl_server_plug_t *a,
     if (a->max_ssf > b->max_ssf) return 1;
     if (a->max_ssf < b->max_ssf) return -1;
 
-    if (SASL_GET_HASH_STRENGTH(a->security_flags) > SASL_GET_HASH_STRENGTH(b->security_flags)) return 1;
-    if (SASL_GET_HASH_STRENGTH(a->security_flags) < SASL_GET_HASH_STRENGTH(b->security_flags)) return -1;
-
     return 0;
 }
 
@@ -425,8 +422,8 @@ int sasl_server_add_plugin(const char *plugname,
 			   sasl_server_plug_init_t *p)
 {
     int plugcount;
-    sasl_server_plug_t *pluglist = NULL;
-    sasl_server_plug_init_t *entry_point = NULL;
+    sasl_server_plug_t *pluglist;
+    sasl_server_plug_init_t *entry_point;
     int result;
     int version;
     int lupe;
@@ -443,8 +440,7 @@ int sasl_server_add_plugin(const char *plugname,
     if ((result != SASL_OK) && (result != SASL_NOUSER)
         && (result != SASL_CONTINUE)) {
 	_sasl_log(NULL, SASL_LOG_DEBUG,
-		  "%s_client_plug_init() failed in sasl_server_add_plugin(): %z\n",
-		  plugname, result);
+		  "server add_plugin entry_point error %z\n", result);
 	return result;
     }
 
@@ -453,8 +449,7 @@ int sasl_server_add_plugin(const char *plugname,
     {
 	_sasl_log(NULL,
 		  SASL_LOG_ERR,
-		  "version mismatch on  sasl_server_add_plugin for '%s': %d expected, but %d reported",
-		  plugname,
+		  "version mismatch on plugin: %d expected, but %d reported",
 		  SASL_SERVER_PLUG_VERSION,
 		  version);
 	return SASL_BADVERS;
@@ -569,31 +564,25 @@ static int server_done(void) {
   global_callbacks.callbacks = NULL;
   global_callbacks.appname = NULL;
 
-  sasl_config_done();
-
   return SASL_OK;
 }
 
 static int server_idle(sasl_conn_t *conn)
 {
-    sasl_server_conn_t *s_conn = NULL;
+    sasl_server_conn_t *s_conn = (sasl_server_conn_t *) conn;
     mechanism_t *m;
 
     if (! mechlist) {
 	return 0;
     }
 
-    if (!conn)
-        return 1;
-    s_conn = (sasl_server_conn_t *) conn;
-
     for (m = s_conn->mech_list;
 	 m != NULL;
 	 m = m->next) {
 	if (m->m.plug->idle
 	    &&  m->m.plug->idle(m->m.plug->glob_context,
 				conn,
-				s_conn->sparams)) {
+				conn ? ((sasl_server_conn_t *)conn)->sparams : NULL)) {
 	    return 1;
 	}
     }
@@ -657,7 +646,7 @@ static int load_config(const sasl_callback_t *verifyfile_cb)
             goto done;
         }
 
-        snprintf(config_filename, len, "%.*s%c%s.conf", (int)path_len, path_to_config,
+        snprintf(config_filename, len, "%.*s%c%s.conf", path_len, path_to_config, 
 	        HIER_DELIMITER, global_callbacks.appname);
 
         /* Ask the application if it's safe to use this file */
@@ -1157,8 +1146,7 @@ int sasl_server_new(const char *service,
 		  tail = serverconn->mech_list;
 	      }
 	      else {
-		  if (tail)
-                     tail->next = new;
+		  tail->next = new;
 		  tail = new;
 	      }
 	      serverconn->mech_length++;
@@ -1434,9 +1422,9 @@ int sasl_server_start(sasl_conn_t *conn,
     }
 
     if (m->m.condition == SASL_CONTINUE) {
-	sasl_server_plug_init_t *entry_point = NULL;
+	sasl_server_plug_init_t *entry_point;
 	void *library = NULL;
-	sasl_server_plug_t *pluglist = NULL;
+	sasl_server_plug_t *pluglist;
 	int version, plugcount;
 	int l = 0;
 
