@@ -1,7 +1,7 @@
 /* SRP SASL plugin
  * Ken Murchison
  * Tim Martin  3/17/00
- * $Id: srp.c,v 1.55 2004/01/12 19:20:00 rjs3 Exp $
+ * $Id: srp.c,v 1.59 2010/11/30 11:41:47 mel Exp $
  */
 /* 
  * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
@@ -87,6 +87,7 @@ typedef unsigned short uint32;
 /* for digest and cipher support */
 #include <openssl/evp.h>
 #include <openssl/hmac.h>
+#include <openssl/md5.h>
 
 #include <sasl.h>
 #define MD5_H  /* suppress internal MD5 */
@@ -100,7 +101,7 @@ typedef unsigned short uint32;
 
 /*****************************  Common Section  *****************************/
 
-static const char plugin_id[] = "$Id: srp.c,v 1.55 2004/01/12 19:20:00 rjs3 Exp $";
+static const char plugin_id[] = "$Id: srp.c,v 1.59 2010/11/30 11:41:47 mel Exp $";
 
 /* Size limit of cipher block size */
 #define SRP_MAXBLOCKSIZE 16
@@ -1854,7 +1855,7 @@ static int srp_server_mech_step1(context_t *text,
     
     /* this will trigger the getting of the aux properties */
     result = params->canon_user(params->utils->conn,
-				user, 0, SASL_CU_AUTHID, oparams);
+				text->authid, 0, SASL_CU_AUTHID, oparams);
     if (result != SASL_OK) goto cleanup;
     
     result = params->canon_user(params->utils->conn,
@@ -1917,7 +1918,7 @@ static int srp_server_mech_step1(context_t *text,
 	    goto cleanup;
 	}
 	
-	result = CalculateV(text, &text->N, &text->g, user,
+	result = CalculateV(text, &text->N, &text->g, text->authid,
 			    auxprop_values[1].values[0], len,
 			    &text->v, &text->salt, &text->saltlen);
 	if (result) {
@@ -2160,12 +2161,16 @@ static int srp_server_mech_step(void *conn_context,
 	|| !oparams)
 	return SASL_BADPARAM;
     
-    sparams->utils->log(NULL, SASL_LOG_DEBUG,
-			"SRP server step %d\n", text->state);
-    
     *serverout = NULL;
     *serveroutlen = 0;
-	
+
+    if (text == NULL) {
+	return SASL_BADPROT;
+    }
+
+    sparams->utils->log(NULL, SASL_LOG_DEBUG,
+			"SRP server step %d\n", text->state);
+
     switch (text->state) {
 
     case 1:
@@ -2197,6 +2202,7 @@ static int srp_setpass(void *glob_context __attribute__((unused)),
 {
     int r;
     char *user = NULL;
+    char *user_only = NULL;
     char *realm = NULL;
     sasl_secret_t *sec = NULL;
     struct propctx *propctx = NULL;
@@ -2210,14 +2216,22 @@ static int srp_setpass(void *glob_context __attribute__((unused)),
 	return SASL_NOMECH;
     }
     
-    r = _plug_parseuser(sparams->utils, &user, &realm, sparams->user_realm,
+    /* NB: Ideally we need to canonicalize userstr here */
+    r = _plug_parseuser(sparams->utils, &user_only, &realm, sparams->user_realm,
 			sparams->serverFQDN, userstr);
+
     if (r) {
 	sparams->utils->seterror(sparams->utils->conn, 0, 
 				 "Error parsing user");
 	return r;
     }
-    
+
+    r = _plug_make_fulluser(sparams->utils, &user, user_only, realm);
+
+    if (r) {
+	goto end;
+    }
+
     if ((flags & SASL_SET_DISABLE) || pass == NULL) {
 	sec = NULL;
     } else {
@@ -2247,7 +2261,8 @@ static int srp_setpass(void *glob_context __attribute__((unused)),
 				     "Error calculating N and g");
 	    goto end;
 	}
-	
+
+	/* user is a full username here */
 	r = CalculateV(text, &N, &g, user, pass, passlen, &v, &salt, &saltlen);
 	if (r) {
 	    sparams->utils->seterror(sparams->utils->conn, 0, 
@@ -2321,6 +2336,7 @@ static int srp_setpass(void *glob_context __attribute__((unused)),
   cleanup:
     
     if (user) 	_plug_free_string(sparams->utils, &user);
+    if (user_only) 	_plug_free_string(sparams->utils, &user_only);
     if (realm) 	_plug_free_string(sparams->utils, &realm);
     if (sec)    _plug_free_secret(sparams->utils, &sec);
     
