@@ -31,14 +31,15 @@
  * crypt(3) based passwd file validation
  * END SYNOPSIS */
 
+#ifdef __GNUC__
+#ident "$Id: auth_getpwent.c,v 1.9 2009/02/13 14:23:26 mel Exp $"
+#endif
+
 /* PUBLIC DEPENDENCIES */
 #include "mechanisms.h"
 #include <unistd.h>
 #include <string.h>
 #include <pwd.h>
-#include <errno.h>
-#include <syslog.h>
-#include <stdio.h>
 
 #ifdef HAVE_CRYPT_H
 #include <crypt.h>
@@ -54,8 +55,6 @@
 #   include <des.h>
 #  endif /* WITH_SSL_DES */
 # endif /* WITH_DES */
-
-# include "globals.h"
 /* END PUBLIC DEPENDENCIES */
 
 #define RETURN(x) return strdup(x)
@@ -74,46 +73,19 @@ auth_getpwent (
 {
     /* VARIABLES */
     struct passwd *pw;			/* pointer to passwd file entry */
-    char *crpt_passwd;			/* encrypted password */
-    int errnum;
     /* END VARIABLES */
   
-    errno = 0;
     pw = getpwnam(login);
-    errnum = errno;
     endpwent();
 
     if (pw == NULL) {
-	if (errnum != 0) {
-	    char *errstr;
-
-	    if (flags & VERBOSE) {
-		syslog(LOG_DEBUG, "DEBUG: auth_getpwent: getpwnam(%s) failure: %m", login);
-	    }
-	    if (asprintf(&errstr, "NO Username lookup failure: %s", strerror(errno)) == -1) {
-		/* XXX the hidden strdup() will likely fail and return NULL here.... */
-		RETURN("NO Username lookup failure: unknown error (ENOMEM formatting strerror())");
-	    }
-	    return errstr;
-	} else {
-	    if (flags & VERBOSE) {
-		syslog(LOG_DEBUG, "DEBUG: auth_getpwent: getpwnam(%s): invalid username", login);
-	    }
-	    RETURN("NO Invalid username");
-	}
+	RETURN("NO");
     }
 
-    crpt_passwd = crypt(password, pw->pw_passwd);
-    if (!crpt_passwd || strcmp(pw->pw_passwd, (const char *)crpt_passwd)) {
-	if (flags & VERBOSE) {
-	    syslog(LOG_DEBUG, "DEBUG: auth_getpwent: %s: invalid password", login);
-	}
-	RETURN("NO Incorrect password");
+    if (strcmp(pw->pw_passwd, (const char *)crypt(password, pw->pw_passwd))) {
+	RETURN("NO");
     }
 
-    if (flags & VERBOSE) {
-	syslog(LOG_DEBUG, "DEBUG: auth_getpwent: OK: %s", login);
-    }
     RETURN("OK");
 }
 
