@@ -53,7 +53,7 @@
  * END SYNOPSIS */
 
 #ifdef __GNUC__
-#ident "$Id: auth_rimap.c,v 1.11 2004/06/30 14:49:28 rjs3 Exp $"
+#ident "$Id: auth_rimap.c,v 1.13 2008/01/23 19:54:54 murch Exp $"
 #endif
 
 /* PUBLIC DEPENDENCIES */
@@ -162,6 +162,7 @@ qstring (
     num_quotes = 0;
     p1 = s;
     while ((p1 = strchr(p1, '"')) != NULL) {
+	p1++;
 	num_quotes++;
     }
     
@@ -438,7 +439,7 @@ auth_rimap (
 	syslog(LOG_WARNING, "auth_rimap: writev: %m");
 	memset(qlogin, 0, strlen(qlogin));
 	free(qlogin);
-	memset(qpass, 0, strlen(qlogin));
+	memset(qpass, 0, strlen(qpass));
 	free(qpass);
 	(void)close(s);
 	return strdup(RESP_IERROR);
@@ -447,7 +448,7 @@ auth_rimap (
     /* don't need these any longer */
     memset(qlogin, 0, strlen(qlogin));
     free(qlogin);
-    memset(qpass, 0, strlen(qlogin));
+    memset(qpass, 0, strlen(qpass));
     free(qpass);
 
     /* read and parse the LOGIN response */
@@ -481,7 +482,7 @@ auth_rimap (
     }
     syslog(LOG_WARNING, "auth_rimap: unexpected response to auth request: %s",
 	   rbuf);
-    return RESP_UNEXPECTED;
+    return strdup(RESP_UNEXPECTED);
     
 }
 
