@@ -55,7 +55,7 @@
 #define PACKAGE "cyrus-sasl"
 
 /* Our version */
-#define VERSION "2.1.20"
+#define VERSION "2.1.25"
 
 /* Visual Studio supports prototypes */
 #define PROTOTYPES     1
@@ -79,8 +79,8 @@ typedef int		    intptr_t;
 
 /* Registry key that contains the locations of the plugins */
 #define SASL_ROOT_KEY "SOFTWARE\\Carnegie Mellon\\Project Cyrus\\SASL Library"
-#define SASL_KEY_V1 SASL_ROOT_KEY "\\Available Plugins"
-#define SASL_PATH_SUBKEY "SearchPath"
+#define SASL_PLUGIN_PATH_ATTR "SearchPath"
+#define SASL_CONF_PATH_ATTR "ConfFile"
 
 /* : This should probably be replaced with a call to a function
    : that gets the proper value from Registry */
@@ -95,11 +95,12 @@ typedef int		    intptr_t;
 #define STATIC_ANONYMOUS 1
 #define STATIC_CRAMMD5 1
 #define STATIC_DIGESTMD5 1
+#define STATIC_SCRAM 1
 #define STATIC_GSSAPIV2 1
 /* #undef STATIC_KERBEROS4 */
 #define STATIC_LOGIN 1
 /* #undef STATIC_MYSQL */
-/* #undef STATIC_OTP */
+#define STATIC_OTP 1
 #define STATIC_PLAIN 1
 #define STATIC_SASLDB 1
 #define STATIC_SRP 1
@@ -110,8 +111,8 @@ typedef int		    intptr_t;
  */
 #define HAVE_MEMCPY 1
 
-#define SASL_PATH_ENV_VAR "SASL_PATH"
 #define PLUGINDIR "C:\\CMU\\bin\\sasl2"
+#define CONFIGDIR "C:\\CMU\\bin\\sasl2"
 
 /* Windows calls these functions something else
  */
@@ -138,6 +139,25 @@ typedef int		    intptr_t;
 typedef int socklen_t;
 #endif /* HAVE_SOCKLEN_T */
 
+/* If we expect to run on XP and later, we have IPv6 support natively */
+#if TARGET_WIN_SYSTEM >= 51
+#if !defined(_WIN32_WINNT)
+/* This forces the inclusion of OS supported functions, with no fallback */
+#define _WIN32_WINNT	0x0510
+#endif
+#endif
+
+#if defined(_MSC_VER) && (_MSC_VER >= 1300)
+/* The following two defines will prevent our own definitions below */
+#define HAVE_GETADDRINFO
+#define HAVE_GETNAMEINFO
+#define HAVE_STRUCT_SOCKADDR_STORAGE
+/* Unless _WIN32_WINNT > 0x0500, Ws2tcpip.h will try to find OS provided
+   getaddrinfo at runtime. It will fallback to Microsoft emulation,
+   if not found */
+#include <Ws2tcpip.h>
+#endif
+
 #if !defined(HAVE_STRUCT_SOCKADDR_STORAGE) && !defined(_SS_MAXSIZE)
 #define	_SS_MAXSIZE	128	/* Implementation specific max size */
 #define	_SS_PADSIZE	(_SS_MAXSIZE - sizeof (struct sockaddr))
@@ -157,14 +177,34 @@ struct sockaddr_storage {
 #ifndef HAVE_GETADDRINFO
 #define	getaddrinfo	sasl_getaddrinfo
 #define	freeaddrinfo	sasl_freeaddrinfo
-#define	getnameinfo	sasl_getnameinfo
 #define	gai_strerror	sasl_gai_strerror
+#endif
+
+#ifndef HAVE_GETNAMEINFO
+#define	getnameinfo	sasl_getnameinfo
+#endif
+
+#if !defined(HAVE_GETNAMEINFO) || !defined(HAVE_GETADDRINFO)
 #include "gai.h"
 #endif
 
+#ifndef AI_NUMERICHOST   /* support glibc 2.0.x */
+#define AI_NUMERICHOST  4
+#define NI_NUMERICHOST  2
+#define NI_NAMEREQD     4
+#define NI_NUMERICSERV  8
+#endif
+
 #include <time.h>
 
-typedef int ssize_t;
+/* Keep in sync with SleepyCat definitions */
+typedef int int32_t;
+typedef __int64 int64_t;
+#ifdef _WIN64
+typedef int64_t ssize_t;
+#else
+typedef int32_t ssize_t;
+#endif
 
 #define HIER_DELIMITER '\\'
 
